require 'gpgme'

require 'perfect_world/db'
require 'perfect_world/error'
require 'perfect_world/storage'
require 'perfect_world/clipboard'

module PerfectWorld

  class Cli

    # Default config.
    CONFIG = {
      'clipboard' => false,
      'database' => File.expand_path('~/.pwm.yml.gpg'),
      'force' => false,
      'length' => 64
    }

    # Sets the config.
    def initialize(config = {})
      @config = CONFIG.merge(config)
      GPGME::Engine.home_dir = @config['gpgdir'] if @config.key?('gpgdir')
    end

    # Loads the database and executes the provided commands.
    def run(commands)
      DB.open(@config.fetch('database'), @config['owner']) do |db|
        commands.each { |cmd, arg| send(cmd, db, arg) }
      end
    end

    private

    # Saves db to another location.
    def backup(db, path)
      db.save(path, @config['owner'])
    end

    # Prints the path to the currently used database.
    def database(_ = nil, _ = nil)
      print_config('database')
    end

    # Deletes a password.
    def delete(db, id)
      if db.delete(id)
        puts "Deleted your #{id} password."
      else
        password_not_found(id)
      end
    end

    # Generats a new password and sends it to stdout.
    def generate(db, id)
      len = @config.fetch('length')
      pwd = @config['force'] ? db.generate!(id, len) : db.generate(id, len)
      print_or_copy_to_clipboard(id, pwd)
    rescue Error => e
      raise Error, "Couldn't create password: #{e.message}"
    end

    # Fetches the password and sends it to stdout.
    def get(db, id)
      print_or_copy_to_clipboard(id, db.fetch(id))
    rescue KeyError
      password_not_found(id)
    end

    # Prints the currently used password length.
    def length(_ = nil, _ = nil)
      print_config('length')
    end

    # Lists all passwords.
    def list(db, _ = nil)
      db.each { |id, password| puts "#{password}   #{id}" }
    end

    # Prints the current key owner.
    def owner(_ = nil, _ = nil)
      puts Storage.find_key(@config['owner']).email
    end

    # Raises an error with a proper message.
    def password_not_found(id)
      raise Error, "Couldn't find your #{id} password."
    end

    # Prints a password or copies it to the clipboard.
    def print_or_copy_to_clipboard(id, password)
      if @config.fetch('clipboard')
        Clipboard << password
        puts "Copied your #{id} password to the clipboard."
      else
        puts password
      end
    end

    # Prints a config entry.
    def print_config(key)
      puts @config[key]
    end
  end
end
