require 'perfect_world/error'
require 'perfect_world/random'
require 'perfect_world/storage'

module PerfectWorld

  # Handles the password database.
  class DB

    def self.open(path, owner = nil)
      db = File.exist?(path) ? load(path) : new({})

      if block_given?
        yield(db)
        db.save(path, owner) if db.changed?
      else
        db
      end
    end

    # Loads the database.
    def self.load(path)
      passwords = Storage.load(path)

      valid?(passwords) ? new(passwords) : raise(CorruptedDatabase, path)
    rescue Error => e
      raise Error, "Couldn't load database: #{e.message}"
    end

    # Checks if a password hash is valid. A valid object is a hash with
    # string values.
    #
    # Returns true if it is valid else false.
    def self.valid?(data)
      data.is_a?(Hash) && data.values.all? { |v| v.is_a?(String) }
    end

    # Inits the db and sets the initial passwords.
    def initialize(passwords = {})
      @passwords = passwords
      @changed = false
    end

    # Generates a new password and puts it in the database.
    #
    #   store.generate(:google, 32)
    #   #=> "B6m/![)A%fqw,\\ti-d`4\"&0>gl+>$0$Z"
    #
    #   store[:google]
    #   #=> "B6m/![)A%fqw,\\ti-d`4\"&0>gl+>$0$Z"
    #
    # Returns the new password or raises an error if the password is already
    # in the database.
    def generate(id, len = 64)
      if ! @passwords.key?(id)
        generate!(id, len)
      else
        raise Error, "Your #{id} password is already in the database."
      end
    end

    # Does the same as DB#generate, but overrides existing passwords.
    def generate!(id, len = 64)
      @passwords[id] = Random.string(len)
      @changed = true
      @passwords[id]
    end

    # Returns the password for the id or nil, if not found.
    def [](id)
      @passwords[id]
    end

    # Fetches a password from the database. It behaves like Hash#fetch.
    def fetch(id, *default)
      @passwords.fetch(id, *default)
    end

    # Deletes the password from the database.
    #
    #   store.delete(:google)
    #   #=> "B6m/![)A%fqw,\\ti-d`4\"&0>gl+>$0$Z"
    #
    #   store[:google]
    #   #=> nil
    #
    # Returns nil if not found.
    def delete(id)
      if (password = @passwords.delete(id))
        @changed = true
        password
      end
    end

    # Returns true, when the database has changed, else false.
    def changed?
      @changed
    end

    # Iterates over all passwords.
    #
    #   store.each do |id, password|
    #     puts "#{password}   #{id}"
    #   end
    #
    # Returns an enumerator, if no block is given.
    def each(&block)
      @passwords.each(&block)
    end

    # Encryptes the database and writes it to disk.
    def save(path, owner = nil)
      Storage.dump(@passwords, owner, path)
    rescue Error => e
      raise Error, "Couldn't save database: #{e.message}"
    end
  end
end
