require 'yaml'
require 'gpgme'
require 'perfect_world/error'

module PerfectWorld

  class Storage

    class << self

      # Loads data from an encrypted file.
      #
      #   PerfectWorld::Storage.load('path/to/file.yml.gpg')  #=> [1, 2]
      #
      # Returns the data.
      def load(path)
        open(path) do |file|
          deserialize(decrypt(file))
        end
      end

      # Writes data into an encrypted file.
      #
      #   PerfectWorld::Storage.dump([1, 2], 'gpguid', /path/to/file.yml.gpg')
      #
      # Raises PerfectWorld::Error on error.
      def dump(data, recipient, path)
        commit(encrypt(serialize(data), find_key(recipient)), path)
      end

      # Finds a secret key in the key ring.
      #
      # Returns GPGME::Key or raises PerfectWorld::KeyNotFound.
      def find_key(term = nil)
        GPGME::Key.find(:secrect, term).first || raise(KeyNotFound, term)
      end

      private

      def open(path)
        File.open(path, 'r') do |f|
          f.flock(File::LOCK_SH)
          yield(f)
        end
      rescue SystemCallError, IOError => e
        raise Error, e.message
      end

      def commit(data, path)
        File.open(path, File::RDWR|File::CREAT, 0600) do |f|
          f.flock(File::LOCK_EX)
          f.truncate(0)
          f.write(data)
        end
      rescue SystemCallError, IOError => e
        raise Error, e.message
      end

      def serialize(data)
        YAML.dump(data)
      rescue => e
        raise Error, e.message
      end

      def deserialize(data)
        YAML.load(data)
      rescue => e
        raise Error, e.message
      end

      def encrypt(data, key)
        GPGME::Crypto.new.encrypt(data, recipients: key, sign: true, signers: key)
      rescue GPGME::Error => e
        raise Error, e.message
      end

      def decrypt(data)
        GPGME::Crypto.new.decrypt(data) do |signature|
          raise(CorruptedDatabase, 'Invalid signature.') unless signature.valid?
        end.to_s
      rescue GPGME::Error => e
        raise Error, e.message
      end
    end
  end
end
